/*
 * A filesystem object accessor.
 *
 * Copyright (C) 2018, 2019, 2020 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/re/c/util/cap_alloc.h>
#include <l4/re/env.h>
#include <l4/sys/factory.h>
#include <l4/sys/semaphore.h>

#include <sys/stat.h>
#include <time.h>

#include <e2access/image.h>
#include <systypes/stat.h>

#include "fs_accessor.h"



/* Initialise an accessor. */

Fs_accessor::Fs_accessor(ext2_filsys fs, ext2_ino_t ino_file, int flags,
                         Paging *paging, size_t size)
: PagingAccessor(paging, size), _fs(fs), _ino_file(ino_file), _flags(flags)
{
  /* Initialise the size from the file. */

  struct ext2_inode inode_file;

  if (!ext2fs_read_inode(fs, ino_file, &inode_file))
  {
    set_size(EXT2_I_SIZE(&inode_file));
  }

  _file = NULL;
}



/* Specific internal methods for ext2fs access. */

size_t Fs_accessor::read_into(char *buffer, size_t to_transfer)
{
  /* Perform potentially many read operations if the buffer is
     larger than the pieces returned by the read function. */

  size_t total = 0;
  unsigned int read;

  while (to_transfer)
  {
    if (ext2fs_file_read(_file, buffer, to_transfer, &read))
      return 0;

    if (!read)
      break;

    to_transfer -= read;
    total += read;
  }

  return total;
}

size_t Fs_accessor::write_from(char *buffer, size_t to_transfer)
{
  /* Perform potentially many write operations if the buffer is
     larger than the pieces consumed by the write function. */

  size_t total = 0;
  unsigned int written;

  while (to_transfer)
  {
    if (ext2fs_file_write(_file, buffer, to_transfer, &written))
      return 0;

    if (!written)
      break;

    to_transfer -= written;
    total += written;
  }

  return total;
}



/* Generic accessor methods implemented for ext2fs. */

unsigned int Fs_accessor::attach()
{
  unsigned int attached = PagingAccessor::attach();

  if (!attached)
  {
    /* Open the file when the first attach operation occurs. */
    /* NOTE: Should handle failure appropriately. */

    if (_file == NULL)
      ext2fs_file_open(_fs, _ino_file, _flags, &_file);

    /* Create a semaphore and make the accessor initially available. */

    _semaphore = l4re_util_cap_alloc();

    if (l4_is_valid_cap(_semaphore))
    {
      l4_factory_create(l4re_env()->factory, L4_PROTO_SEMAPHORE, _semaphore);
      l4_semaphore_up(_semaphore);
    }
  }

  return attached;
}

unsigned int Fs_accessor::detach()
{
  unsigned int attached = PagingAccessor::detach();

  if (!attached)
  {
    if (_file != NULL)
    {
      ext2fs_file_flush(_file);
      ext2fs_file_close(_file);
      _file = NULL;
    }

    /* Release the semaphore. */

    if (l4_is_valid_cap(_semaphore))
      l4re_util_cap_free_um(_semaphore);
  }

  return attached;
}

long Fs_accessor::flush()
{
  long err;

  l4_semaphore_down(_semaphore, L4_IPC_NEVER);

  err = ext2fs_file_flush(_file);

  l4_semaphore_up(_semaphore);

  return err;
}

size_t Fs_accessor::fstat(char *buffer, size_t length)
{
  size_t nread;
  struct stat st;
  sys_stat_t *statbuf = (sys_stat_t *) buffer;

  if (length < sizeof(sys_stat_t))
    return 0;

  l4_semaphore_down(_semaphore, L4_IPC_NEVER);

  if (image_stat_inode(_fs, _ino_file, &st))
    nread = 0;
  else
  {
    nread = sizeof(sys_stat_t);
    *statbuf = (sys_stat_t)
               {st.st_dev, st.st_ino, st.st_mode, st.st_nlink,
                st.st_uid, st.st_gid, st.st_rdev, st.st_size,
                {st.st_atim.tv_sec, st.st_atim.tv_nsec},
                {st.st_mtim.tv_sec, st.st_mtim.tv_nsec},
                {st.st_ctim.tv_sec, st.st_ctim.tv_nsec},
                st.st_blksize, st.st_blocks};
  }

  l4_semaphore_up(_semaphore);

  return nread;
}

size_t Fs_accessor::read(size_t start, char *buffer, size_t length)
{
  size_t nread;

  l4_semaphore_down(_semaphore, L4_IPC_NEVER);

  if (ext2fs_file_lseek(_file, start, EXT2_SEEK_SET, 0))
    nread = 0;
  else
    nread = read_into(buffer, length);

  l4_semaphore_up(_semaphore);

  return nread;
}

size_t Fs_accessor::write(size_t start, char *buffer, size_t length)
{
  size_t written;

  l4_semaphore_down(_semaphore, L4_IPC_NEVER);

  if (ext2fs_file_lseek(_file, start, EXT2_SEEK_SET, 0))
    written = 0;
  else
    written = write_from(buffer, length);

  l4_semaphore_up(_semaphore);

  return written;
}
